;; This Bigloo module is intended to provide word hyphenation.
;; The program is an implementation of the algorithm created
;; by Frank Liang for Donald Knuth's TeX typographic system and
;; exposed in Liang's PhD dissertation available on the TeX Users
;; Group site here: http://www.tug.org/docs/liang/
;; Most of this implementation is borrowed from Phil Bewig's work
;; available here: http://sites.google.com/site/schemephil/, along
;; with his paper describing the program.
;;
;; word hy-phen-a-tion by com-pu-ter

(module hyphenation
   (import patterns call-hyphen)
   (export (hyphenate lang word)))

(define (fold-left op base xs)
  (if (null? xs)
      base
      (fold-left op (op base (car xs)) (cdr xs))))

(define (build-list n x)  ;; LB make-list caused name clash with Bigloo
  (let loop ((n n) (xs '()))
    (if (zero? n) xs
      (loop (- n 1) (cons x xs)))))

(define a-null '())

(define (a-look c a)
  (cond ((null? a) #f)
        ((char-ci<? c (caar a)) #f)
        ((char-ci<? (caar a) c) (a-look c (cdr a)))
        (else (car a))))

(define (a-bind c x a)
  (cond ((null? a) (list (cons c x)))
        ((char-ci<? c (caar a)) (cons (cons c x) a))
        ((char-ci<? (caar a) c) (cons (car a) (a-bind c x (cdr a))))
        (else (cons (cons c x) (cdr a)))))

(define t-null (cons '() a-null))

(define (t-look ks t)
  (if (null? ks)
      (if (pair? (car t)) (caar t) #f)
      (let ((x (a-look (car ks) (cdr t))))
        (if x (t-look (cdr ks) (cdr x)) #f))))

(define (t-bind ks x t)
  (if (null? ks) (cons (list x) (cdr t))
    (let* ((a1 (a-look (car ks) (cdr t)))
           (t2 (t-bind (cdr ks) x (if (pair? a1) (cdr a1) t-null))))
      (cons (car t) (a-bind (car ks) t2 (cdr t))))))

(define (split-pat pat)
;; Modif. LB: patterns are no more symbols, but strings
  (let loop ((ps (string->list pat)) (cs '()) (ns '())) ;; LB
    (cond ((null? ps)
            (if (= (length cs) (length ns))
                (values (reverse cs) (reverse (cons 0 ns)))
                (values (reverse cs) (reverse ns))))
          ((and (null? (cdr ps)) (char-numeric? (car ps)))
            (values (reverse cs)
                    (reverse (cons (- (char->integer (car ps)) 48) ns))))
          ((not (char-numeric? (car ps)))
            (loop (cdr ps) (cons (car ps) cs) (cons 0 ns)))
          (else (loop (cddr ps)
                      (cons (cadr ps) cs)
                      (cons (- (char->integer (car ps)) 48) ns))))))

(define (split-exn exn)
;; Modif. LB: patterns are no more symbols, but strings
  (let loop ((ps (append (list #\.) (string->list exn) (list #\.))) ;; LB
             (cs '()) (ns '()))
    (cond ((null? ps) (values (reverse cs) (reverse (cons 6 ns))))
          ((char=? (car ps) #\-)
            (loop (cddr ps) (cons (cadr ps) cs) (cons 7 ns)))
          (else (loop (cdr ps) (cons (car ps) cs) (cons 6 ns))))))

(define (select-hyphens lang)
   (let ((hyphens-nb (vector-length hyphens)))
      (do ((i 0 (+ i 1)))
	  ((or
	    (eq? i hyphens-nb)
	    (eq? lang (vector-ref (vector-ref hyphens i) 0)))
	   (if (< i hyphens-nb)
	       (vector (vector-ref (vector-ref hyphens i) 1)
		       (vector-ref (vector-ref hyphens i) 2))
	       #f)))))

(define (make-t-hyph lang)
;; Modif. LB: patterns and exceptions are now in the separate module patterns
   (let ((hyphens (select-hyphens lang)))
      (if hyphens
	  (let ((patterns (vector-ref hyphens 0))    ;; LB
		(exceptions (vector-ref hyphens 1))) ;; LB
	     (let loop ((pats patterns) (exns exceptions) (t t-null))
		(cond
		   ((null? exns) t)
		   ((null? pats)
		    (call-with-values (lambda () (split-exn (car exns)))
				      (lambda (cs ns) (loop pats (cdr exns) (t-bind cs ns t)))))
		   (else (call-with-values (lambda () (split-pat (car pats)))
					   (lambda (cs ns) (loop (cdr pats) exns (t-bind cs ns t))))))))
	  #f)))

(define (t-looks ks t)
  (if (null? ks)
      (if (pair? (car t)) (list (caar t)) '())
      (let ((x (a-look (car ks) (cdr t))))
        (if x
            (if (pair? (car t))
                (cons (caar t) (t-looks (cdr ks) (cdr x)))
                (t-looks (cdr ks) (cdr x)))
            (if (pair? (car t)) (list (caar t)) '())))))

(define (max-rule xs ys)
  (let loop ((xs xs) (ys ys) (zs '()))
    (if (or (null? xs) (null? ys))
        (append (reverse zs) xs)
        (loop (cdr xs) (cdr ys) (cons (max (car xs) (car ys)) zs)))))

(define (fixup xs)
  (cons 0 (cons 0 (cdddr (reverse (cons 0 (cons 0 (cddr xs))))))))

(define (hyphenate lang word)
   (let ((t-hyph (make-t-hyph lang)))
      (if t-hyph
	  (if (< (string-length word) 5) (list word)
	      (let loop ((ws (append (list #\.) (string->list word) (list #\.)))
			 (front (build-list (+ (string-length word) 3) 0)) (back '()))
		 (if (null? ws)
		     (let loop ((cs (string->list word)) (hs (fixup back)) (p '()) (ps '()))
			(cond ((null? (cdr hs)) (reverse (cons (list->string (reverse p)) ps)))
			      ((odd? (car hs))
			       (loop (cdr cs) (cdr hs) (list (car cs))
				     (cons (list->string (reverse p)) ps)))
			      (else (loop (cdr cs) (cdr hs) (cons (car cs) p) ps))))
		     (let ((new-front (fold-left max-rule front (t-looks ws t-hyph))))
			(loop (cdr ws) (cdr new-front) (cons (car new-front) back))))))
	  (list word)))) ;; We don't know how to hyphenate, no patterns for this language
